-- This reservoir simulation deck is made available under the Open Database
-- License: http://opendatacommons.org/licenses/odbl/1.0/. Any rights in
-- individual contents of the database are licensed under the Database Contents
-- License: http://opendatacommons.org/licenses/dbcl/1.0/

-- Copyright (C) 2015 Statoil

RGIR
1 /
RGPR
1 /
RGIPG
1 /
RGIPL
1 /

RGIT
1 /
RGPT
1 /
RWIR
1 /
RWPR
1 /
RWPT
1 /
RWIP
1 /
ROPR
1 /
ROIR
1 /
ROIT
1 /
ROIPG
1 /
ROIPL
1 /

ROPT
1 /
