-- This reservoir simulation deck is made available under the Open Database
-- License: http://opendatacommons.org/licenses/odbl/1.0/. Any rights in
-- individual contents of the database are licensed under the Database Contents
-- License: http://opendatacommons.org/licenses/dbcl/1.0/

-- Copyright (C) 2015 Statoil

EQUALS
-- FIPs based on geological layers
FIPGL 1    1  46   1 112   1  3   /  Garn 
FIPGL 2    1  46   1 112   4  4   /  Not
FIPGL 3    1  46   1 112   5  5   /  Ile 2.2 
FIPGL 4    1  46   1 112   6  8   /  Ile 2.1
FIPGL 5    1  46   1 112   9  11  /  Ile 1
FIPGL 6    1  46   1 112   12 12  /  Tofte 2.2
FIPGL 7    1  46   1 112   13 15  /  Tofte 2.1
FIPGL 8    1  46   1 112   16 18  /  Tofte 1
FIPGL 9    1  46   1 112   19 22  /  Tilje

-- FIPs based on numerical layers
FIPNL 1    1  46   1 112   1  1   /  Garn 3
FIPNL 2    1  46   1 112   2  2   /  Garn 2
FIPNL 3    1  46   1 112   3  3   /  Garn 1w
FIPNL 4    1  46   1 112   4  4   /  Not   
FIPNL 5    1  46   1 112   5  5   /  Ile 2.2
FIPNL 6    1  46   1 112   6  6   /  Ile 2.1.3      
FIPNL 7    1  46   1 112   7  7   /  Ile 2.1.2
FIPNL 8    1  46   1 112   8  8   /  Ile 2.1.1
FIPNL 9    1  46   1 112   9  9   /  Ile 1.3                  
FIPNL 10   1  46   1 112   10 10  /  Ile 1.2      
FIPNL 11   1  46   1 112   11 11  /  Ile 1.1
FIPNL 12   1  46   1 112   12 12  /  Tofte 2.2            
FIPNL 13   1  46   1 112   13 13  /  Tofte 2.1.3
FIPNL 14   1  46   1 112   14 14  /  Tofte 2.1.2
FIPNL 15   1  46   1 112   15 15  /  Tofte 2.1.1
FIPNL 16   1  46   1 112   16 16  /  Tofte 1.2.2
FIPNL 17   1  46   1 112   17 17  /  Tofte 1.2.1
FIPNL 18   1  46   1 112   18 18  /  Tofte 1.1
FIPNL 19   1  46   1 112   19 19  /  Tilje 4
FIPNL 20   1  46   1 112   20 20  /  Tilje 3
FIPNL 21   1  46   1 112   21 21  /  Tilje 2
FIPNL 22   1  46   1 112   22 22  /  Tilje 1
/