from math import sin
import random

def run():
    pass

print("sin(0) = {}".format(sin(0)))
#---
if random.random() > 0.25:
    print("Large outcome")
else:
    print("Small result")
A = 100
B = A / 10
C = B * 20
