from typing import Any, ClassVar, Dict, Iterator, List, Tuple

from typing import overload
import datetime
import numpy
C0nn: eclArrType
CHAR: eclArrType
DOUB: eclArrType
EDIT: eclSectionType
EModel_add_filter_docstring: str
GRID: eclSectionType
INTE: eclArrType
LOGI: eclArrType
MESS: eclArrType
PROPS: eclSectionType
REAL: eclArrType
REGIONS: eclSectionType
RUNSPEC: eclSectionType
SCHEDULE: eclSectionType
SOLUTION: eclSectionType
SUMMARY: eclSectionType

current_ecl_state: EclipseState
current_report_step: int
current_schedule: Schedule
current_summary_state: SummaryState

class Builtin:
    def __init__(self) -> None: ...
    def __getitem__(self, index) -> Any: ...
    @property
    def ACF(self) -> Any: ...
    @property
    def ACTCO2S(self) -> Any: ...
    @property
    def ACTDIMS(self) -> Any: ...
    @property
    def ACTION(self) -> Any: ...
    @property
    def ACTIONG(self) -> Any: ...
    @property
    def ACTIONR(self) -> Any: ...
    @property
    def ACTIONS(self) -> Any: ...
    @property
    def ACTIONW(self) -> Any: ...
    @property
    def ACTIONX(self) -> Any: ...
    @property
    def ACTNUM(self) -> Any: ...
    @property
    def ACTPARAM(self) -> Any: ...
    @property
    def ADD(self) -> Any: ...
    @property
    def ADDREG(self) -> Any: ...
    @property
    def ADDZCORN(self) -> Any: ...
    @property
    def ADSALNOD(self) -> Any: ...
    @property
    def ADSORP(self) -> Any: ...
    @property
    def AITS(self) -> Any: ...
    @property
    def AITSOFF(self) -> Any: ...
    @property
    def ALKADS(self) -> Any: ...
    @property
    def ALKALINE(self) -> Any: ...
    @property
    def ALKROCK(self) -> Any: ...
    @property
    def ALL(self) -> Any: ...
    @property
    def ALPOLADS(self) -> Any: ...
    @property
    def ALSURFAD(self) -> Any: ...
    @property
    def ALSURFST(self) -> Any: ...
    @property
    def AMALGAM(self) -> Any: ...
    @property
    def API(self) -> Any: ...
    @property
    def APIGROUP(self) -> Any: ...
    @property
    def APILIM(self) -> Any: ...
    @property
    def APIVD(self) -> Any: ...
    @property
    def AQANCONL(self) -> Any: ...
    @property
    def AQANNC(self) -> Any: ...
    @property
    def AQANTRC(self) -> Any: ...
    @property
    def AQUALIST(self) -> Any: ...
    @property
    def AQUANCON(self) -> Any: ...
    @property
    def AQUCHGAS(self) -> Any: ...
    @property
    def AQUCHWAT(self) -> Any: ...
    @property
    def AQUCON(self) -> Any: ...
    @property
    def AQUCT(self) -> Any: ...
    @property
    def AQUCWFAC(self) -> Any: ...
    @property
    def AQUDIMS(self) -> Any: ...
    @property
    def AQUFET(self) -> Any: ...
    @property
    def AQUFETP(self) -> Any: ...
    @property
    def AQUFLUX(self) -> Any: ...
    @property
    def AQUIFER_PROBE_ANALYTIC(self) -> Any: ...
    @property
    def AQUIFER_PROBE_ANALYTIC_NAMED(self) -> Any: ...
    @property
    def AQUIFER_PROBE_NUMERIC(self) -> Any: ...
    @property
    def AQUNNC(self) -> Any: ...
    @property
    def AQUNUM(self) -> Any: ...
    @property
    def AQUTAB(self) -> Any: ...
    @property
    def AUTOCOAR(self) -> Any: ...
    @property
    def AUTOREF(self) -> Any: ...
    @property
    def BC(self) -> Any: ...
    @property
    def BCCON(self) -> Any: ...
    @property
    def BCPROP(self) -> Any: ...
    @property
    def BDENSITY(self) -> Any: ...
    @property
    def BGGI(self) -> Any: ...
    @property
    def BIC(self) -> Any: ...
    @property
    def BIGMODEL(self) -> Any: ...
    @property
    def BIOFILM(self) -> Any: ...
    @property
    def BIOFPARA(self) -> Any: ...
    @property
    def BIOTCOEF(self) -> Any: ...
    @property
    def BLACKOIL(self) -> Any: ...
    @property
    def BLOCK_PROBE(self) -> Any: ...
    @property
    def BLOCK_PROBE300(self) -> Any: ...
    @property
    def BLOCK_PROBE900(self) -> Any: ...
    @property
    def BOGI(self) -> Any: ...
    @property
    def BOUNDARY(self) -> Any: ...
    @property
    def BOX(self) -> Any: ...
    @property
    def BPARA(self) -> Any: ...
    @property
    def BPIDIMS(self) -> Any: ...
    @property
    def BRANPROP(self) -> Any: ...
    @property
    def BRINE(self) -> Any: ...
    @property
    def BTOBALFA(self) -> Any: ...
    @property
    def BTOBALFV(self) -> Any: ...
    @property
    def CALTRAC(self) -> Any: ...
    @property
    def CARFIN(self) -> Any: ...
    @property
    def CART(self) -> Any: ...
    @property
    def CBMOPTS(self) -> Any: ...
    @property
    def CECON(self) -> Any: ...
    @property
    def CECONT(self) -> Any: ...
    @property
    def CIRCLE(self) -> Any: ...
    @property
    def CNAMES(self) -> Any: ...
    @property
    def CO2SOL(self) -> Any: ...
    @property
    def CO2STOR(self) -> Any: ...
    @property
    def CO2STORE(self) -> Any: ...
    @property
    def COAL(self) -> Any: ...
    @property
    def COALADS(self) -> Any: ...
    @property
    def COALNUM(self) -> Any: ...
    @property
    def COALPP(self) -> Any: ...
    @property
    def COARSEN(self) -> Any: ...
    @property
    def COLLAPSE(self) -> Any: ...
    @property
    def COLUMNS(self) -> Any: ...
    @property
    def COMPDAT(self) -> Any: ...
    @property
    def COMPDATX(self) -> Any: ...
    @property
    def COMPFLSH(self) -> Any: ...
    @property
    def COMPIMB(self) -> Any: ...
    @property
    def COMPINJK(self) -> Any: ...
    @property
    def COMPLMPL(self) -> Any: ...
    @property
    def COMPLUMP(self) -> Any: ...
    @property
    def COMPOFF(self) -> Any: ...
    @property
    def COMPORD(self) -> Any: ...
    @property
    def COMPRIV(self) -> Any: ...
    @property
    def COMPRP(self) -> Any: ...
    @property
    def COMPRPL(self) -> Any: ...
    @property
    def COMPS(self) -> Any: ...
    @property
    def COMPSEGL(self) -> Any: ...
    @property
    def COMPSEGS(self) -> Any: ...
    @property
    def COMPTRAJ(self) -> Any: ...
    @property
    def COMPVE(self) -> Any: ...
    @property
    def COMPVEL(self) -> Any: ...
    @property
    def CONNECTION_PROBE(self) -> Any: ...
    @property
    def CONNECTION_PROBE_OPM(self) -> Any: ...
    @property
    def COORD(self) -> Any: ...
    @property
    def COORDSYS(self) -> Any: ...
    @property
    def COPY(self) -> Any: ...
    @property
    def COPYBOX(self) -> Any: ...
    @property
    def COPYREG(self) -> Any: ...
    @property
    def CPIFACT(self) -> Any: ...
    @property
    def CPIFACTL(self) -> Any: ...
    @property
    def CPR(self) -> Any: ...
    @property
    def CREF(self) -> Any: ...
    @property
    def CREFW(self) -> Any: ...
    @property
    def CREFWS(self) -> Any: ...
    @property
    def CRITPERM(self) -> Any: ...
    @property
    def CSKIN(self) -> Any: ...
    @property
    def DATE(self) -> Any: ...
    @property
    def DATES(self) -> Any: ...
    @property
    def DATUM(self) -> Any: ...
    @property
    def DATUMR(self) -> Any: ...
    @property
    def DATUMRX(self) -> Any: ...
    @property
    def DCQDEFN(self) -> Any: ...
    @property
    def DEBUG_(self) -> Any: ...
    @property
    def DELAYACT(self) -> Any: ...
    @property
    def DENAQA(self) -> Any: ...
    @property
    def DENSITY(self) -> Any: ...
    @property
    def DEPTH(self) -> Any: ...
    @property
    def DEPTHTAB(self) -> Any: ...
    @property
    def DEPTHZ(self) -> Any: ...
    @property
    def DIAGDISP(self) -> Any: ...
    @property
    def DIFF(self) -> Any: ...
    @property
    def DIFFAGAS(self) -> Any: ...
    @property
    def DIFFAWAT(self) -> Any: ...
    @property
    def DIFFC(self) -> Any: ...
    @property
    def DIFFCGAS(self) -> Any: ...
    @property
    def DIFFCOAL(self) -> Any: ...
    @property
    def DIFFCWAT(self) -> Any: ...
    @property
    def DIFFDP(self) -> Any: ...
    @property
    def DIFFMICP(self) -> Any: ...
    @property
    def DIFFMMF(self) -> Any: ...
    @property
    def DIFFMR(self) -> Any: ...
    @property
    def DIFFMTHT(self) -> Any: ...
    @property
    def DIFFMX(self) -> Any: ...
    @property
    def DIFFMY(self) -> Any: ...
    @property
    def DIFFMZ(self) -> Any: ...
    @property
    def DIFFR(self) -> Any: ...
    @property
    def DIFFTHT(self) -> Any: ...
    @property
    def DIFFUSE(self) -> Any: ...
    @property
    def DIFFX(self) -> Any: ...
    @property
    def DIFFY(self) -> Any: ...
    @property
    def DIFFZ(self) -> Any: ...
    @property
    def DIMENS(self) -> Any: ...
    @property
    def DIMPES(self) -> Any: ...
    @property
    def DIMPLICT(self) -> Any: ...
    @property
    def DISGAS(self) -> Any: ...
    @property
    def DISGASW(self) -> Any: ...
    @property
    def DISPDIMS(self) -> Any: ...
    @property
    def DISPERC(self) -> Any: ...
    @property
    def DISPERSE(self) -> Any: ...
    @property
    def DOMAINS(self) -> Any: ...
    @property
    def DPGRID(self) -> Any: ...
    @property
    def DPKRMOD(self) -> Any: ...
    @property
    def DPNUM(self) -> Any: ...
    @property
    def DR(self) -> Any: ...
    @property
    def DREF(self) -> Any: ...
    @property
    def DREFS(self) -> Any: ...
    @property
    def DRILPRI(self) -> Any: ...
    @property
    def DRSDT(self) -> Any: ...
    @property
    def DRSDTCON(self) -> Any: ...
    @property
    def DRSDTR(self) -> Any: ...
    @property
    def DRV(self) -> Any: ...
    @property
    def DRVDT(self) -> Any: ...
    @property
    def DRVDTR(self) -> Any: ...
    @property
    def DSPDEINT(self) -> Any: ...
    @property
    def DTHETA(self) -> Any: ...
    @property
    def DTHETAV(self) -> Any: ...
    @property
    def DUALPERM(self) -> Any: ...
    @property
    def DUALPORO(self) -> Any: ...
    @property
    def DUMPCUPL(self) -> Any: ...
    @property
    def DUMPFLUX(self) -> Any: ...
    @property
    def DX(self) -> Any: ...
    @property
    def DXV(self) -> Any: ...
    @property
    def DY(self) -> Any: ...
    @property
    def DYNAMICR(self) -> Any: ...
    @property
    def DYNRDIMS(self) -> Any: ...
    @property
    def DYV(self) -> Any: ...
    @property
    def DZ(self) -> Any: ...
    @property
    def DZMATRIX(self) -> Any: ...
    @property
    def DZMTRX(self) -> Any: ...
    @property
    def DZMTRXV(self) -> Any: ...
    @property
    def DZNET(self) -> Any: ...
    @property
    def DZV(self) -> Any: ...
    @property
    def ECHO(self) -> Any: ...
    @property
    def ECLMC(self) -> Any: ...
    @property
    def EDIT(self) -> Any: ...
    @property
    def EDITNNC(self) -> Any: ...
    @property
    def EDITNNCR(self) -> Any: ...
    @property
    def EHYSTR(self) -> Any: ...
    @property
    def EHYSTRR(self) -> Any: ...
    @property
    def END(self) -> Any: ...
    @property
    def ENDACTIO(self) -> Any: ...
    @property
    def ENDBOX(self) -> Any: ...
    @property
    def ENDDYN(self) -> Any: ...
    @property
    def ENDFIN(self) -> Any: ...
    @property
    def ENDINC(self) -> Any: ...
    @property
    def ENDNUM(self) -> Any: ...
    @property
    def ENDPOINT_SPECIFIERS(self) -> Any: ...
    @property
    def ENDSCALE(self) -> Any: ...
    @property
    def ENDSKIP(self) -> Any: ...
    @property
    def ENKRVD(self) -> Any: ...
    @property
    def ENPCVD(self) -> Any: ...
    @property
    def ENPTVD(self) -> Any: ...
    @property
    def ENSPCVD(self) -> Any: ...
    @property
    def EOS(self) -> Any: ...
    @property
    def EOSNUM(self) -> Any: ...
    @property
    def EPSDBGS(self) -> Any: ...
    @property
    def EPSDEBUG(self) -> Any: ...
    @property
    def EQLDIMS(self) -> Any: ...
    @property
    def EQLNUM(self) -> Any: ...
    @property
    def EQLOPTS(self) -> Any: ...
    @property
    def EQLZCORN(self) -> Any: ...
    @property
    def EQUALREG(self) -> Any: ...
    @property
    def EQUALS(self) -> Any: ...
    @property
    def EQUIL(self) -> Any: ...
    @property
    def ESSNODE(self) -> Any: ...
    @property
    def EXCAVATE(self) -> Any: ...
    @property
    def EXCEL(self) -> Any: ...
    @property
    def EXIT(self) -> Any: ...
    @property
    def EXTFIN(self) -> Any: ...
    @property
    def EXTHOST(self) -> Any: ...
    @property
    def EXTRAPMS(self) -> Any: ...
    @property
    def EXTREPGL(self) -> Any: ...
    @property
    def FAULTDIM(self) -> Any: ...
    @property
    def FAULTS(self) -> Any: ...
    @property
    def FBHPDEF(self) -> Any: ...
    @property
    def FHERCHBL(self) -> Any: ...
    @property
    def FIELD(self) -> Any: ...
    @property
    def FIELDSEP(self) -> Any: ...
    @property
    def FIELD_PROBE(self) -> Any: ...
    @property
    def FIELD_PROBE_OPM(self) -> Any: ...
    @property
    def FILEUNIT(self) -> Any: ...
    @property
    def FILLEPS(self) -> Any: ...
    @property
    def FIPNUM(self) -> Any: ...
    @property
    def FIPOWG(self) -> Any: ...
    @property
    def FIPSEP(self) -> Any: ...
    @property
    def FIP_PROBE(self) -> Any: ...
    @property
    def FLUXNUM(self) -> Any: ...
    @property
    def FLUXREG(self) -> Any: ...
    @property
    def FLUXTYPE(self) -> Any: ...
    @property
    def FMTHMD(self) -> Any: ...
    @property
    def FMTIN(self) -> Any: ...
    @property
    def FMTOUT(self) -> Any: ...
    @property
    def FMWSET(self) -> Any: ...
    @property
    def FOAM(self) -> Any: ...
    @property
    def FOAMADS(self) -> Any: ...
    @property
    def FOAMDCYO(self) -> Any: ...
    @property
    def FOAMDCYW(self) -> Any: ...
    @property
    def FOAMFCN(self) -> Any: ...
    @property
    def FOAMFRM(self) -> Any: ...
    @property
    def FOAMFSC(self) -> Any: ...
    @property
    def FOAMFSO(self) -> Any: ...
    @property
    def FOAMFST(self) -> Any: ...
    @property
    def FOAMFSW(self) -> Any: ...
    @property
    def FOAMMOB(self) -> Any: ...
    @property
    def FOAMMOBP(self) -> Any: ...
    @property
    def FOAMMOBS(self) -> Any: ...
    @property
    def FOAMOPTS(self) -> Any: ...
    @property
    def FOAMROCK(self) -> Any: ...
    @property
    def FORMFEED(self) -> Any: ...
    @property
    def FRICTION(self) -> Any: ...
    @property
    def FULLIMP(self) -> Any: ...
    @property
    def GAS(self) -> Any: ...
    @property
    def GASBEGIN(self) -> Any: ...
    @property
    def GASCONC(self) -> Any: ...
    @property
    def GASDENT(self) -> Any: ...
    @property
    def GASEND(self) -> Any: ...
    @property
    def GASFCOMP(self) -> Any: ...
    @property
    def GASFDECR(self) -> Any: ...
    @property
    def GASFDELC(self) -> Any: ...
    @property
    def GASFIELD(self) -> Any: ...
    @property
    def GASFTARG(self) -> Any: ...
    @property
    def GASJT(self) -> Any: ...
    @property
    def GASMONTH(self) -> Any: ...
    @property
    def GASPERIO(self) -> Any: ...
    @property
    def GASSATC(self) -> Any: ...
    @property
    def GASVISCT(self) -> Any: ...
    @property
    def GASWAT(self) -> Any: ...
    @property
    def GASYEAR(self) -> Any: ...
    @property
    def GCALECON(self) -> Any: ...
    @property
    def GCOMPIDX(self) -> Any: ...
    @property
    def GCONCAL(self) -> Any: ...
    @property
    def GCONENG(self) -> Any: ...
    @property
    def GCONINJE(self) -> Any: ...
    @property
    def GCONPRI(self) -> Any: ...
    @property
    def GCONPROD(self) -> Any: ...
    @property
    def GCONSALE(self) -> Any: ...
    @property
    def GCONSUMP(self) -> Any: ...
    @property
    def GCONTOL(self) -> Any: ...
    @property
    def GCUTBACK(self) -> Any: ...
    @property
    def GCUTBACT(self) -> Any: ...
    @property
    def GCVD(self) -> Any: ...
    @property
    def GDCQ(self) -> Any: ...
    @property
    def GDCQECON(self) -> Any: ...
    @property
    def GDFILE(self) -> Any: ...
    @property
    def GDIMS(self) -> Any: ...
    @property
    def GDORIENT(self) -> Any: ...
    @property
    def GDRILPOT(self) -> Any: ...
    @property
    def GECON(self) -> Any: ...
    @property
    def GECONT(self) -> Any: ...
    @property
    def GEFAC(self) -> Any: ...
    @property
    def GETDATA(self) -> Any: ...
    @property
    def GETGLOB(self) -> Any: ...
    @property
    def GI(self) -> Any: ...
    @property
    def GIALL(self) -> Any: ...
    @property
    def GIMODEL(self) -> Any: ...
    @property
    def GINODE(self) -> Any: ...
    @property
    def GLIFTLIM(self) -> Any: ...
    @property
    def GLIFTOPT(self) -> Any: ...
    @property
    def GMWSET(self) -> Any: ...
    @property
    def GNETDP(self) -> Any: ...
    @property
    def GNETINJE(self) -> Any: ...
    @property
    def GNETPUMP(self) -> Any: ...
    @property
    def GPMAINT(self) -> Any: ...
    @property
    def GRADGRUP(self) -> Any: ...
    @property
    def GRADRESV(self) -> Any: ...
    @property
    def GRADRFT(self) -> Any: ...
    @property
    def GRADWELL(self) -> Any: ...
    @property
    def GRAVCONS(self) -> Any: ...
    @property
    def GRAVDR(self) -> Any: ...
    @property
    def GRAVDRB(self) -> Any: ...
    @property
    def GRAVDRM(self) -> Any: ...
    @property
    def GRAVITY(self) -> Any: ...
    @property
    def GRDREACH(self) -> Any: ...
    @property
    def GRID(self) -> Any: ...
    @property
    def GRIDFILE(self) -> Any: ...
    @property
    def GRIDOPTS(self) -> Any: ...
    @property
    def GRIDUNIT(self) -> Any: ...
    @property
    def GROUP_PROBE(self) -> Any: ...
    @property
    def GROUP_PROBE_OPM(self) -> Any: ...
    @property
    def GRUPMAST(self) -> Any: ...
    @property
    def GRUPNET(self) -> Any: ...
    @property
    def GRUPRIG(self) -> Any: ...
    @property
    def GRUPSLAV(self) -> Any: ...
    @property
    def GRUPTARG(self) -> Any: ...
    @property
    def GRUPTREE(self) -> Any: ...
    @property
    def GSATINJE(self) -> Any: ...
    @property
    def GSATPROD(self) -> Any: ...
    @property
    def GSEPCOND(self) -> Any: ...
    @property
    def GSF(self) -> Any: ...
    @property
    def GSSCPTST(self) -> Any: ...
    @property
    def GSWINGF(self) -> Any: ...
    @property
    def GTADD(self) -> Any: ...
    @property
    def GTMULT(self) -> Any: ...
    @property
    def GUIDECAL(self) -> Any: ...
    @property
    def GUIDERAT(self) -> Any: ...
    @property
    def GUPFREQ(self) -> Any: ...
    @property
    def GWRTWCV(self) -> Any: ...
    @property
    def H2SOL(self) -> Any: ...
    @property
    def H2STORE(self) -> Any: ...
    @property
    def HALFTRAN(self) -> Any: ...
    @property
    def HAxxxxxx(self) -> Any: ...
    @property
    def HBNUM(self) -> Any: ...
    @property
    def HDISP(self) -> Any: ...
    @property
    def HEATCR(self) -> Any: ...
    @property
    def HEATCRT(self) -> Any: ...
    @property
    def HMAQUCT(self) -> Any: ...
    @property
    def HMAQUFET(self) -> Any: ...
    @property
    def HMAQUNUM(self) -> Any: ...
    @property
    def HMDIMS(self) -> Any: ...
    @property
    def HMFAULTS(self) -> Any: ...
    @property
    def HMMLAQUN(self) -> Any: ...
    @property
    def HMMLCTAQ(self) -> Any: ...
    @property
    def HMMLFTAQ(self) -> Any: ...
    @property
    def HMMLTWCN(self) -> Any: ...
    @property
    def HMMULTFT(self) -> Any: ...
    @property
    def HMMULTSG(self) -> Any: ...
    @property
    def HMMULTxx(self) -> Any: ...
    @property
    def HMPROPS(self) -> Any: ...
    @property
    def HMROCK(self) -> Any: ...
    @property
    def HMROCKT(self) -> Any: ...
    @property
    def HMRREF(self) -> Any: ...
    @property
    def HMWELCON(self) -> Any: ...
    @property
    def HMWPIMLT(self) -> Any: ...
    @property
    def HMxxxxxx(self) -> Any: ...
    @property
    def HRFIN(self) -> Any: ...
    @property
    def HWELLS(self) -> Any: ...
    @property
    def HWKRO(self) -> Any: ...
    @property
    def HWKRORG(self) -> Any: ...
    @property
    def HWKRORW(self) -> Any: ...
    @property
    def HWKRW(self) -> Any: ...
    @property
    def HWKRWR(self) -> Any: ...
    @property
    def HWPCW(self) -> Any: ...
    @property
    def HWSNUM(self) -> Any: ...
    @property
    def HWSOGCR(self) -> Any: ...
    @property
    def HWSOWCR(self) -> Any: ...
    @property
    def HWSWCR(self) -> Any: ...
    @property
    def HWSWL(self) -> Any: ...
    @property
    def HWSWLPC(self) -> Any: ...
    @property
    def HWSWU(self) -> Any: ...
    @property
    def HXFIN(self) -> Any: ...
    @property
    def HYDRHEAD(self) -> Any: ...
    @property
    def HYFIN(self) -> Any: ...
    @property
    def HYMOBGDR(self) -> Any: ...
    @property
    def HYST(self) -> Any: ...
    @property
    def HYSTCHCK(self) -> Any: ...
    @property
    def HZFIN(self) -> Any: ...
    @property
    def IHOST(self) -> Any: ...
    @property
    def IMBNUM(self) -> Any: ...
    @property
    def IMBNUMMF(self) -> Any: ...
    @property
    def IMKRVD(self) -> Any: ...
    @property
    def IMPCVD(self) -> Any: ...
    @property
    def IMPES(self) -> Any: ...
    @property
    def IMPLICIT(self) -> Any: ...
    @property
    def IMPORT(self) -> Any: ...
    @property
    def IMPTVD(self) -> Any: ...
    @property
    def IMSPCVD(self) -> Any: ...
    @property
    def INCLUDE(self) -> Any: ...
    @property
    def INIT(self) -> Any: ...
    @property
    def INRAD(self) -> Any: ...
    @property
    def INSPEC(self) -> Any: ...
    @property
    def INTPC(self) -> Any: ...
    @property
    def IONROCK(self) -> Any: ...
    @property
    def IONXROCK(self) -> Any: ...
    @property
    def IONXSURF(self) -> Any: ...
    @property
    def IPCG(self) -> Any: ...
    @property
    def IPCW(self) -> Any: ...
    @property
    def ISGCR(self) -> Any: ...
    @property
    def ISGL(self) -> Any: ...
    @property
    def ISGLPC(self) -> Any: ...
    @property
    def ISGU(self) -> Any: ...
    @property
    def ISOGCR(self) -> Any: ...
    @property
    def ISOLNUM(self) -> Any: ...
    @property
    def ISOWCR(self) -> Any: ...
    @property
    def ISWCR(self) -> Any: ...
    @property
    def ISWL(self) -> Any: ...
    @property
    def ISWLPC(self) -> Any: ...
    @property
    def ISWU(self) -> Any: ...
    @property
    def JFUNC(self) -> Any: ...
    @property
    def JFUNCR(self) -> Any: ...
    @property
    def KRNUM(self) -> Any: ...
    @property
    def KRNUMMF(self) -> Any: ...
    @property
    def LAB(self) -> Any: ...
    @property
    def LANGMPL(self) -> Any: ...
    @property
    def LANGMUIR(self) -> Any: ...
    @property
    def LANGSOLV(self) -> Any: ...
    @property
    def LCUNIT(self) -> Any: ...
    @property
    def LGR(self) -> Any: ...
    @property
    def LGRCOPY(self) -> Any: ...
    @property
    def LGRFREE(self) -> Any: ...
    @property
    def LGRLOCK(self) -> Any: ...
    @property
    def LGROFF(self) -> Any: ...
    @property
    def LGRON(self) -> Any: ...
    @property
    def LICENSES(self) -> Any: ...
    @property
    def LIFTOPT(self) -> Any: ...
    @property
    def LINCOM(self) -> Any: ...
    @property
    def LINKPERM(self) -> Any: ...
    @property
    def LIVEOIL(self) -> Any: ...
    @property
    def LKRO(self) -> Any: ...
    @property
    def LKRORG(self) -> Any: ...
    @property
    def LKRORW(self) -> Any: ...
    @property
    def LKRW(self) -> Any: ...
    @property
    def LKRWR(self) -> Any: ...
    @property
    def LOAD(self) -> Any: ...
    @property
    def LOWSALT(self) -> Any: ...
    @property
    def LPCW(self) -> Any: ...
    @property
    def LSALTFNC(self) -> Any: ...
    @property
    def LSLTWNUM(self) -> Any: ...
    @property
    def LSNUM(self) -> Any: ...
    @property
    def LSOGCR(self) -> Any: ...
    @property
    def LSOWCR(self) -> Any: ...
    @property
    def LSWCR(self) -> Any: ...
    @property
    def LSWL(self) -> Any: ...
    @property
    def LSWLPC(self) -> Any: ...
    @property
    def LSWU(self) -> Any: ...
    @property
    def LTOSIGMA(self) -> Any: ...
    @property
    def LWKRO(self) -> Any: ...
    @property
    def LWKRORG(self) -> Any: ...
    @property
    def LWKRORW(self) -> Any: ...
    @property
    def LWKRW(self) -> Any: ...
    @property
    def LWKRWR(self) -> Any: ...
    @property
    def LWPCW(self) -> Any: ...
    @property
    def LWSLTNUM(self) -> Any: ...
    @property
    def LWSNUM(self) -> Any: ...
    @property
    def LWSOGCR(self) -> Any: ...
    @property
    def LWSOWCR(self) -> Any: ...
    @property
    def LWSWCR(self) -> Any: ...
    @property
    def LWSWL(self) -> Any: ...
    @property
    def LWSWLPC(self) -> Any: ...
    @property
    def LWSWU(self) -> Any: ...
    @property
    def LX(self) -> Any: ...
    @property
    def LXFIN(self) -> Any: ...
    @property
    def LY(self) -> Any: ...
    @property
    def LYFIN(self) -> Any: ...
    @property
    def LZ(self) -> Any: ...
    @property
    def LZFIN(self) -> Any: ...
    @property
    def MAPAXES(self) -> Any: ...
    @property
    def MAPUNITS(self) -> Any: ...
    @property
    def MASSFLOW(self) -> Any: ...
    @property
    def MATCORR(self) -> Any: ...
    @property
    def MAXVALUE(self) -> Any: ...
    @property
    def MECH(self) -> Any: ...
    @property
    def MEMORY(self) -> Any: ...
    @property
    def MESSAGE(self) -> Any: ...
    @property
    def MESSAGES(self) -> Any: ...
    @property
    def MESSOPTS(self) -> Any: ...
    @property
    def MESSSRVC(self) -> Any: ...
    @property
    def METRIC(self) -> Any: ...
    @property
    def MICP(self) -> Any: ...
    @property
    def MINNNCT(self) -> Any: ...
    @property
    def MINNPCOL(self) -> Any: ...
    @property
    def MINPORV(self) -> Any: ...
    @property
    def MINPV(self) -> Any: ...
    @property
    def MINPVV(self) -> Any: ...
    @property
    def MINVALUE(self) -> Any: ...
    @property
    def MISC(self) -> Any: ...
    @property
    def MISCIBLE(self) -> Any: ...
    @property
    def MISCNUM(self) -> Any: ...
    @property
    def MLANG(self) -> Any: ...
    @property
    def MLANGSLV(self) -> Any: ...
    @property
    def MONITOR(self) -> Any: ...
    @property
    def MPFANUM(self) -> Any: ...
    @property
    def MPFNNC(self) -> Any: ...
    @property
    def MSFN(self) -> Any: ...
    @property
    def MSGFILE(self) -> Any: ...
    @property
    def MSUM_PROBE(self) -> Any: ...
    @property
    def MULSGGD(self) -> Any: ...
    @property
    def MULSGGDV(self) -> Any: ...
    @property
    def MULTFLT(self) -> Any: ...
    @property
    def MULTIN(self) -> Any: ...
    @property
    def MULTIPLY(self) -> Any: ...
    @property
    def MULTIREG(self) -> Any: ...
    @property
    def MULTNUM(self) -> Any: ...
    @property
    def MULTOUT(self) -> Any: ...
    @property
    def MULTOUTS(self) -> Any: ...
    @property
    def MULTPV(self) -> Any: ...
    @property
    def MULTREAL(self) -> Any: ...
    @property
    def MULTREGD(self) -> Any: ...
    @property
    def MULTREGH(self) -> Any: ...
    @property
    def MULTREGP(self) -> Any: ...
    @property
    def MULTREGT(self) -> Any: ...
    @property
    def MULTSIG(self) -> Any: ...
    @property
    def MULTSIGV(self) -> Any: ...
    @property
    def MULT_XYZ(self) -> Any: ...
    @property
    def MW(self) -> Any: ...
    @property
    def MWS(self) -> Any: ...
    @property
    def NARROW(self) -> Any: ...
    @property
    def NCOMPS(self) -> Any: ...
    @property
    def NCONSUMP(self) -> Any: ...
    @property
    def NEFAC(self) -> Any: ...
    @property
    def NETBALAN(self) -> Any: ...
    @property
    def NETCOMPA(self) -> Any: ...
    @property
    def NETWORK(self) -> Any: ...
    @property
    def NETWORK_PROBE(self) -> Any: ...
    @property
    def NEWTRAN(self) -> Any: ...
    @property
    def NEXTSTEP(self) -> Any: ...
    @property
    def NEXTSTPL(self) -> Any: ...
    @property
    def NINENUM(self) -> Any: ...
    @property
    def NINEPOIN(self) -> Any: ...
    @property
    def NMATOPTS(self) -> Any: ...
    @property
    def NMATRIX(self) -> Any: ...
    @property
    def NMESSAGE(self) -> Any: ...
    @property
    def NNC(self) -> Any: ...
    @property
    def NNEWTF(self) -> Any: ...
    @property
    def NOCASC(self) -> Any: ...
    @property
    def NODEPROP(self) -> Any: ...
    @property
    def NODPPM(self) -> Any: ...
    @property
    def NOECHO(self) -> Any: ...
    @property
    def NOGGF(self) -> Any: ...
    @property
    def NOGRAV(self) -> Any: ...
    @property
    def NOHMD(self) -> Any: ...
    @property
    def NOHMO(self) -> Any: ...
    @property
    def NOHYST(self) -> Any: ...
    @property
    def NOINSPEC(self) -> Any: ...
    @property
    def NOMONITO(self) -> Any: ...
    @property
    def NONNC(self) -> Any: ...
    @property
    def NORSSPEC(self) -> Any: ...
    @property
    def NOSIM(self) -> Any: ...
    @property
    def NOWARN(self) -> Any: ...
    @property
    def NOWARNEP(self) -> Any: ...
    @property
    def NRSOUT(self) -> Any: ...
    @property
    def NSTACK(self) -> Any: ...
    @property
    def NTG(self) -> Any: ...
    @property
    def NUMRES(self) -> Any: ...
    @property
    def NUPCOL(self) -> Any: ...
    @property
    def NWATREM(self) -> Any: ...
    @property
    def NXFIN(self) -> Any: ...
    @property
    def NYFIN(self) -> Any: ...
    @property
    def NZFIN(self) -> Any: ...
    @property
    def OCOMPIDX(self) -> Any: ...
    @property
    def OFM(self) -> Any: ...
    @property
    def OIL(self) -> Any: ...
    @property
    def OILAPI(self) -> Any: ...
    @property
    def OILCOMPR(self) -> Any: ...
    @property
    def OILDENT(self) -> Any: ...
    @property
    def OILJT(self) -> Any: ...
    @property
    def OILMW(self) -> Any: ...
    @property
    def OILVISCT(self) -> Any: ...
    @property
    def OILVTIM(self) -> Any: ...
    @property
    def OLDTRAN(self) -> Any: ...
    @property
    def OLDTRANR(self) -> Any: ...
    @property
    def OPERATE(self) -> Any: ...
    @property
    def OPERATER(self) -> Any: ...
    @property
    def OPERNUM(self) -> Any: ...
    @property
    def OPTIONS(self) -> Any: ...
    @property
    def OPTIONS3(self) -> Any: ...
    @property
    def OUTRAD(self) -> Any: ...
    @property
    def OUTSOL(self) -> Any: ...
    @property
    def OVERBURD(self) -> Any: ...
    @property
    def PARALLEL(self) -> Any: ...
    @property
    def PARAOPTS(self) -> Any: ...
    @property
    def PARTTRAC(self) -> Any: ...
    @property
    def PATHS(self) -> Any: ...
    @property
    def PBUB(self) -> Any: ...
    @property
    def PBVD(self) -> Any: ...
    @property
    def PCFACT(self) -> Any: ...
    @property
    def PCG(self) -> Any: ...
    @property
    def PCG32D(self) -> Any: ...
    @property
    def PCRIT(self) -> Any: ...
    @property
    def PCW(self) -> Any: ...
    @property
    def PCW32D(self) -> Any: ...
    @property
    def PDEW(self) -> Any: ...
    @property
    def PDVD(self) -> Any: ...
    @property
    def PEBI(self) -> Any: ...
    @property
    def PECOEFS(self) -> Any: ...
    @property
    def PEDIMS(self) -> Any: ...
    @property
    def PEGTABX(self) -> Any: ...
    @property
    def PEKTABX(self) -> Any: ...
    @property
    def PENUM(self) -> Any: ...
    @property
    def PERFORMANCE_PROBE(self) -> Any: ...
    @property
    def PERMAVE(self) -> Any: ...
    @property
    def PERMFACT(self) -> Any: ...
    @property
    def PERMJFUN(self) -> Any: ...
    @property
    def PERMR(self) -> Any: ...
    @property
    def PERMTHT(self) -> Any: ...
    @property
    def PERMX(self) -> Any: ...
    @property
    def PERMXY(self) -> Any: ...
    @property
    def PERMY(self) -> Any: ...
    @property
    def PERMYZ(self) -> Any: ...
    @property
    def PERMZ(self) -> Any: ...
    @property
    def PERMZX(self) -> Any: ...
    @property
    def PETGRID(self) -> Any: ...
    @property
    def PETOPTS(self) -> Any: ...
    @property
    def PICOND(self) -> Any: ...
    @property
    def PIMTDIMS(self) -> Any: ...
    @property
    def PIMULTAB(self) -> Any: ...
    @property
    def PINCH(self) -> Any: ...
    @property
    def PINCHNUM(self) -> Any: ...
    @property
    def PINCHOUT(self) -> Any: ...
    @property
    def PINCHREG(self) -> Any: ...
    @property
    def PINCHXY(self) -> Any: ...
    @property
    def PINTDIMS(self) -> Any: ...
    @property
    def PLMIXNUM(self) -> Any: ...
    @property
    def PLMIXPAR(self) -> Any: ...
    @property
    def PLYADS(self) -> Any: ...
    @property
    def PLYADSS(self) -> Any: ...
    @property
    def PLYATEMP(self) -> Any: ...
    @property
    def PLYCAMAX(self) -> Any: ...
    @property
    def PLYDHFLF(self) -> Any: ...
    @property
    def PLYESAL(self) -> Any: ...
    @property
    def PLYKRRF(self) -> Any: ...
    @property
    def PLYMAX(self) -> Any: ...
    @property
    def PLYMWINJ(self) -> Any: ...
    @property
    def PLYOPTS(self) -> Any: ...
    @property
    def PLYRMDEN(self) -> Any: ...
    @property
    def PLYROCK(self) -> Any: ...
    @property
    def PLYROCKM(self) -> Any: ...
    @property
    def PLYSHEAR(self) -> Any: ...
    @property
    def PLYSHLOG(self) -> Any: ...
    @property
    def PLYTRRF(self) -> Any: ...
    @property
    def PLYTRRFA(self) -> Any: ...
    @property
    def PLYVISC(self) -> Any: ...
    @property
    def PLYVISCS(self) -> Any: ...
    @property
    def PLYVISCT(self) -> Any: ...
    @property
    def PLYVMH(self) -> Any: ...
    @property
    def PLYVSCST(self) -> Any: ...
    @property
    def PMAX(self) -> Any: ...
    @property
    def PMISC(self) -> Any: ...
    @property
    def POELCOEF(self) -> Any: ...
    @property
    def POLYMER(self) -> Any: ...
    @property
    def POLYMW(self) -> Any: ...
    @property
    def PORO(self) -> Any: ...
    @property
    def PORV(self) -> Any: ...
    @property
    def PPCWMAX(self) -> Any: ...
    @property
    def PRATIO(self) -> Any: ...
    @property
    def PRECSALT(self) -> Any: ...
    @property
    def PREF(self) -> Any: ...
    @property
    def PREFS(self) -> Any: ...
    @property
    def PRESSURE(self) -> Any: ...
    @property
    def PRIORITY(self) -> Any: ...
    @property
    def PROPS(self) -> Any: ...
    @property
    def PRORDER(self) -> Any: ...
    @property
    def PRVD(self) -> Any: ...
    @property
    def PSTEADY(self) -> Any: ...
    @property
    def PSWRG(self) -> Any: ...
    @property
    def PSWRO(self) -> Any: ...
    @property
    def PVCDO(self) -> Any: ...
    @property
    def PVCO(self) -> Any: ...
    @property
    def PVDG(self) -> Any: ...
    @property
    def PVDO(self) -> Any: ...
    @property
    def PVDS(self) -> Any: ...
    @property
    def PVTG(self) -> Any: ...
    @property
    def PVTGW(self) -> Any: ...
    @property
    def PVTGWO(self) -> Any: ...
    @property
    def PVTNUM(self) -> Any: ...
    @property
    def PVTO(self) -> Any: ...
    @property
    def PVTSOL(self) -> Any: ...
    @property
    def PVTW(self) -> Any: ...
    @property
    def PVTWSALT(self) -> Any: ...
    @property
    def PVT_M(self) -> Any: ...
    @property
    def PVZG(self) -> Any: ...
    @property
    def PYACTION(self) -> Any: ...
    @property
    def PYINPUT(self) -> Any: ...
    @property
    def QDRILL(self) -> Any: ...
    @property
    def QHRATING(self) -> Any: ...
    @property
    def QMOBIL(self) -> Any: ...
    @property
    def RADFIN(self) -> Any: ...
    @property
    def RADFIN4(self) -> Any: ...
    @property
    def RADIAL(self) -> Any: ...
    @property
    def RAINFALL(self) -> Any: ...
    @property
    def RBEDCONT(self) -> Any: ...
    @property
    def RCMASTS(self) -> Any: ...
    @property
    def REACHES(self) -> Any: ...
    @property
    def READDATA(self) -> Any: ...
    @property
    def REFINE(self) -> Any: ...
    @property
    def REGDIMS(self) -> Any: ...
    @property
    def REGION2REGION_PROBE(self) -> Any: ...
    @property
    def REGION2REGION_PROBE_E300(self) -> Any: ...
    @property
    def REGIONS(self) -> Any: ...
    @property
    def REGION_PROBE(self) -> Any: ...
    @property
    def REGION_PROBE_OPM(self) -> Any: ...
    @property
    def RESIDNUM(self) -> Any: ...
    @property
    def RESTART(self) -> Any: ...
    @property
    def RESVNUM(self) -> Any: ...
    @property
    def RHO(self) -> Any: ...
    @property
    def RIVDEBUG(self) -> Any: ...
    @property
    def RIVERSYS(self) -> Any: ...
    @property
    def RIVRDIMS(self) -> Any: ...
    @property
    def RIVRPROP(self) -> Any: ...
    @property
    def RIVRXSEC(self) -> Any: ...
    @property
    def RIVSALT(self) -> Any: ...
    @property
    def RIVTRACE(self) -> Any: ...
    @property
    def RKTRMDIR(self) -> Any: ...
    @property
    def ROCK(self) -> Any: ...
    @property
    def ROCK2D(self) -> Any: ...
    @property
    def ROCK2DTR(self) -> Any: ...
    @property
    def ROCKCOMP(self) -> Any: ...
    @property
    def ROCKFRAC(self) -> Any: ...
    @property
    def ROCKNUM(self) -> Any: ...
    @property
    def ROCKOPTS(self) -> Any: ...
    @property
    def ROCKPAMA(self) -> Any: ...
    @property
    def ROCKTAB(self) -> Any: ...
    @property
    def ROCKTABH(self) -> Any: ...
    @property
    def ROCKTABW(self) -> Any: ...
    @property
    def ROCKTHSG(self) -> Any: ...
    @property
    def ROCKTSIG(self) -> Any: ...
    @property
    def ROCKV(self) -> Any: ...
    @property
    def ROCKWNOD(self) -> Any: ...
    @property
    def RPTCPL(self) -> Any: ...
    @property
    def RPTGRID(self) -> Any: ...
    @property
    def RPTGRIDL(self) -> Any: ...
    @property
    def RPTHMD(self) -> Any: ...
    @property
    def RPTHMG(self) -> Any: ...
    @property
    def RPTHMW(self) -> Any: ...
    @property
    def RPTINIT(self) -> Any: ...
    @property
    def RPTISOL(self) -> Any: ...
    @property
    def RPTONLY(self) -> Any: ...
    @property
    def RPTONLYO(self) -> Any: ...
    @property
    def RPTPROPS(self) -> Any: ...
    @property
    def RPTREGS(self) -> Any: ...
    @property
    def RPTRST(self) -> Any: ...
    @property
    def RPTRUNSP(self) -> Any: ...
    @property
    def RPTSCHED(self) -> Any: ...
    @property
    def RPTSMRY(self) -> Any: ...
    @property
    def RPTSOL(self) -> Any: ...
    @property
    def RS(self) -> Any: ...
    @property
    def RSCONST(self) -> Any: ...
    @property
    def RSCONSTT(self) -> Any: ...
    @property
    def RSGI(self) -> Any: ...
    @property
    def RSSPEC(self) -> Any: ...
    @property
    def RSVD(self) -> Any: ...
    @property
    def RSW(self) -> Any: ...
    @property
    def RTEMP(self) -> Any: ...
    @property
    def RTEMPA(self) -> Any: ...
    @property
    def RTEMPVD(self) -> Any: ...
    @property
    def RUNSPEC(self) -> Any: ...
    @property
    def RUNSUM(self) -> Any: ...
    @property
    def RV(self) -> Any: ...
    @property
    def RVCONST(self) -> Any: ...
    @property
    def RVCONSTT(self) -> Any: ...
    @property
    def RVGI(self) -> Any: ...
    @property
    def RVVD(self) -> Any: ...
    @property
    def RVW(self) -> Any: ...
    @property
    def RVWVD(self) -> Any: ...
    @property
    def RWGSALT(self) -> Any: ...
    @property
    def SALINITY(self) -> Any: ...
    @property
    def SALT(self) -> Any: ...
    @property
    def SALTMF(self) -> Any: ...
    @property
    def SALTNODE(self) -> Any: ...
    @property
    def SALTP(self) -> Any: ...
    @property
    def SALTPVD(self) -> Any: ...
    @property
    def SALTREST(self) -> Any: ...
    @property
    def SALTSOL(self) -> Any: ...
    @property
    def SALTVD(self) -> Any: ...
    @property
    def SAMG(self) -> Any: ...
    @property
    def SATNUM(self) -> Any: ...
    @property
    def SATOPTS(self) -> Any: ...
    @property
    def SAVE(self) -> Any: ...
    @property
    def SBIOF(self) -> Any: ...
    @property
    def SCALC(self) -> Any: ...
    @property
    def SCALECRS(self) -> Any: ...
    @property
    def SCALELIM(self) -> Any: ...
    @property
    def SCDATAB(self) -> Any: ...
    @property
    def SCDETAB(self) -> Any: ...
    @property
    def SCDPDIMS(self) -> Any: ...
    @property
    def SCDPTAB(self) -> Any: ...
    @property
    def SCDPTRAC(self) -> Any: ...
    @property
    def SCHEDULE(self) -> Any: ...
    @property
    def SCVD(self) -> Any: ...
    @property
    def SDENSITY(self) -> Any: ...
    @property
    def SEGMENT_PROBE(self) -> Any: ...
    @property
    def SEPARATE(self) -> Any: ...
    @property
    def SEPVALS(self) -> Any: ...
    @property
    def SFOAM(self) -> Any: ...
    @property
    def SGAS(self) -> Any: ...
    @property
    def SGCR(self) -> Any: ...
    @property
    def SGCWMIS(self) -> Any: ...
    @property
    def SGF32D(self) -> Any: ...
    @property
    def SGFN(self) -> Any: ...
    @property
    def SGL(self) -> Any: ...
    @property
    def SGLPC(self) -> Any: ...
    @property
    def SGOF(self) -> Any: ...
    @property
    def SGOFLET(self) -> Any: ...
    @property
    def SGU(self) -> Any: ...
    @property
    def SGWFN(self) -> Any: ...
    @property
    def SHRATE(self) -> Any: ...
    @property
    def SIGMA(self) -> Any: ...
    @property
    def SIGMAGDV(self) -> Any: ...
    @property
    def SIGMATH(self) -> Any: ...
    @property
    def SIGMAV(self) -> Any: ...
    @property
    def SIMULATE(self) -> Any: ...
    @property
    def SKIP(self) -> Any: ...
    @property
    def SKIP100(self) -> Any: ...
    @property
    def SKIP300(self) -> Any: ...
    @property
    def SKIPREST(self) -> Any: ...
    @property
    def SKPRPOLY(self) -> Any: ...
    @property
    def SKPRWAT(self) -> Any: ...
    @property
    def SKRO(self) -> Any: ...
    @property
    def SKRORG(self) -> Any: ...
    @property
    def SKRORW(self) -> Any: ...
    @property
    def SKRW(self) -> Any: ...
    @property
    def SKRWR(self) -> Any: ...
    @property
    def SLAVES(self) -> Any: ...
    @property
    def SLGOF(self) -> Any: ...
    @property
    def SMICR(self) -> Any: ...
    @property
    def SMRYDIMS(self) -> Any: ...
    @property
    def SMULTX(self) -> Any: ...
    @property
    def SMULTY(self) -> Any: ...
    @property
    def SMULTZ(self) -> Any: ...
    @property
    def SOCRS(self) -> Any: ...
    @property
    def SOF2(self) -> Any: ...
    @property
    def SOF3(self) -> Any: ...
    @property
    def SOF32D(self) -> Any: ...
    @property
    def SOGCR(self) -> Any: ...
    @property
    def SOIL(self) -> Any: ...
    @property
    def SOLID(self) -> Any: ...
    @property
    def SOLUTION(self) -> Any: ...
    @property
    def SOLVCONC(self) -> Any: ...
    @property
    def SOLVDIMS(self) -> Any: ...
    @property
    def SOLVDIRS(self) -> Any: ...
    @property
    def SOLVENT(self) -> Any: ...
    @property
    def SOLVFRAC(self) -> Any: ...
    @property
    def SOLVNUM(self) -> Any: ...
    @property
    def SOMGAS(self) -> Any: ...
    @property
    def SOMWAT(self) -> Any: ...
    @property
    def SORWMIS(self) -> Any: ...
    @property
    def SOURCE(self) -> Any: ...
    @property
    def SOWCR(self) -> Any: ...
    @property
    def SOXYG(self) -> Any: ...
    @property
    def SPECGRID(self) -> Any: ...
    @property
    def SPECHEAT(self) -> Any: ...
    @property
    def SPECROCK(self) -> Any: ...
    @property
    def SPIDER(self) -> Any: ...
    @property
    def SPOLY(self) -> Any: ...
    @property
    def SPOLYMW(self) -> Any: ...
    @property
    def SSFN(self) -> Any: ...
    @property
    def SSGCR(self) -> Any: ...
    @property
    def SSGL(self) -> Any: ...
    @property
    def SSOGCR(self) -> Any: ...
    @property
    def SSOL(self) -> Any: ...
    @property
    def SSOWCR(self) -> Any: ...
    @property
    def SSWCR(self) -> Any: ...
    @property
    def SSWL(self) -> Any: ...
    @property
    def SSWU(self) -> Any: ...
    @property
    def START(self) -> Any: ...
    @property
    def STCOND(self) -> Any: ...
    @property
    def STOG(self) -> Any: ...
    @property
    def STONE(self) -> Any: ...
    @property
    def STONE1(self) -> Any: ...
    @property
    def STONE1EX(self) -> Any: ...
    @property
    def STONE2(self) -> Any: ...
    @property
    def STOW(self) -> Any: ...
    @property
    def STREQUIL(self) -> Any: ...
    @property
    def STRESSEQUILNUM(self) -> Any: ...
    @property
    def STWG(self) -> Any: ...
    @property
    def SUMMARY(self) -> Any: ...
    @property
    def SUMTHIN(self) -> Any: ...
    @property
    def SUREA(self) -> Any: ...
    @property
    def SURF(self) -> Any: ...
    @property
    def SURFACT(self) -> Any: ...
    @property
    def SURFACTW(self) -> Any: ...
    @property
    def SURFADDW(self) -> Any: ...
    @property
    def SURFADS(self) -> Any: ...
    @property
    def SURFCAPD(self) -> Any: ...
    @property
    def SURFESAL(self) -> Any: ...
    @property
    def SURFNUM(self) -> Any: ...
    @property
    def SURFOPTS(self) -> Any: ...
    @property
    def SURFROCK(self) -> Any: ...
    @property
    def SURFST(self) -> Any: ...
    @property
    def SURFSTES(self) -> Any: ...
    @property
    def SURFVISC(self) -> Any: ...
    @property
    def SURFWNUM(self) -> Any: ...
    @property
    def SWAT(self) -> Any: ...
    @property
    def SWATINIT(self) -> Any: ...
    @property
    def SWCR(self) -> Any: ...
    @property
    def SWF32D(self) -> Any: ...
    @property
    def SWFN(self) -> Any: ...
    @property
    def SWINGFAC(self) -> Any: ...
    @property
    def SWL(self) -> Any: ...
    @property
    def SWLPC(self) -> Any: ...
    @property
    def SWOF(self) -> Any: ...
    @property
    def SWOFLET(self) -> Any: ...
    @property
    def SWU(self) -> Any: ...
    @property
    def TABDIMS(self) -> Any: ...
    @property
    def TBLK(self) -> Any: ...
    @property
    def TCRIT(self) -> Any: ...
    @property
    def TEMP(self) -> Any: ...
    @property
    def TEMPI(self) -> Any: ...
    @property
    def TEMPNODE(self) -> Any: ...
    @property
    def TEMPTVD(self) -> Any: ...
    @property
    def TEMPVD(self) -> Any: ...
    @property
    def THCGAS(self) -> Any: ...
    @property
    def THCO2MIX(self) -> Any: ...
    @property
    def THCOIL(self) -> Any: ...
    @property
    def THCONR(self) -> Any: ...
    @property
    def THCONSF(self) -> Any: ...
    @property
    def THCROCK(self) -> Any: ...
    @property
    def THCWATER(self) -> Any: ...
    @property
    def THELCOEF(self) -> Any: ...
    @property
    def THERMAL(self) -> Any: ...
    @property
    def THERMEXR(self) -> Any: ...
    @property
    def THPRES(self) -> Any: ...
    @property
    def THPRESFT(self) -> Any: ...
    @property
    def TIGHTEN(self) -> Any: ...
    @property
    def TIGHTENP(self) -> Any: ...
    @property
    def TIME(self) -> Any: ...
    @property
    def TITLE(self) -> Any: ...
    @property
    def TLMIXPAR(self) -> Any: ...
    @property
    def TLPMIXPA(self) -> Any: ...
    @property
    def TNUM(self) -> Any: ...
    @property
    def TOLCRIT(self) -> Any: ...
    @property
    def TOPS(self) -> Any: ...
    @property
    def TPAMEPS(self) -> Any: ...
    @property
    def TPAMEPSS(self) -> Any: ...
    @property
    def TRACER(self) -> Any: ...
    @property
    def TRACERKM(self) -> Any: ...
    @property
    def TRACERKP(self) -> Any: ...
    @property
    def TRACERS(self) -> Any: ...
    @property
    def TRACITVD(self) -> Any: ...
    @property
    def TRACTVD(self) -> Any: ...
    @property
    def TRADS(self) -> Any: ...
    @property
    def TRANGL(self) -> Any: ...
    @property
    def TRANR(self) -> Any: ...
    @property
    def TRANTHT(self) -> Any: ...
    @property
    def TRANX(self) -> Any: ...
    @property
    def TRANY(self) -> Any: ...
    @property
    def TRANZ(self) -> Any: ...
    @property
    def TRDCY(self) -> Any: ...
    @property
    def TRDIF(self) -> Any: ...
    @property
    def TRDIS(self) -> Any: ...
    @property
    def TREF(self) -> Any: ...
    @property
    def TREFS(self) -> Any: ...
    @property
    def TRKPF(self) -> Any: ...
    @property
    def TRNHD(self) -> Any: ...
    @property
    def TRPLPORO(self) -> Any: ...
    @property
    def TRROCK(self) -> Any: ...
    @property
    def TSTEP(self) -> Any: ...
    @property
    def TUNING(self) -> Any: ...
    @property
    def TUNINGDP(self) -> Any: ...
    @property
    def TUNINGH(self) -> Any: ...
    @property
    def TUNINGL(self) -> Any: ...
    @property
    def TUNINGS(self) -> Any: ...
    @property
    def TVDP(self) -> Any: ...
    @property
    def TZONE(self) -> Any: ...
    @property
    def UDADIMS(self) -> Any: ...
    @property
    def UDQ(self) -> Any: ...
    @property
    def UDQDIMS(self) -> Any: ...
    @property
    def UDQPARAM(self) -> Any: ...
    @property
    def UDT(self) -> Any: ...
    @property
    def UDTDIMS(self) -> Any: ...
    @property
    def UNCODHMD(self) -> Any: ...
    @property
    def UNIFIN(self) -> Any: ...
    @property
    def UNIFOUT(self) -> Any: ...
    @property
    def UNIFOUTS(self) -> Any: ...
    @property
    def UNIFSAVE(self) -> Any: ...
    @property
    def USECUPL(self) -> Any: ...
    @property
    def USEFLUX(self) -> Any: ...
    @property
    def USENOFLO(self) -> Any: ...
    @property
    def VAPOIL(self) -> Any: ...
    @property
    def VAPPARS(self) -> Any: ...
    @property
    def VAPWAT(self) -> Any: ...
    @property
    def VCRIT(self) -> Any: ...
    @property
    def VDFLOW(self) -> Any: ...
    @property
    def VDFLOWR(self) -> Any: ...
    @property
    def VE(self) -> Any: ...
    @property
    def VEDEBUG(self) -> Any: ...
    @property
    def VEFIN(self) -> Any: ...
    @property
    def VEFRAC(self) -> Any: ...
    @property
    def VEFRACP(self) -> Any: ...
    @property
    def VEFRACPV(self) -> Any: ...
    @property
    def VEFRACV(self) -> Any: ...
    @property
    def VFPCHK(self) -> Any: ...
    @property
    def VFPIDIMS(self) -> Any: ...
    @property
    def VFPINJ(self) -> Any: ...
    @property
    def VFPPDIMS(self) -> Any: ...
    @property
    def VFPPROD(self) -> Any: ...
    @property
    def VFPTABL(self) -> Any: ...
    @property
    def VISAGE(self) -> Any: ...
    @property
    def VISCAQA(self) -> Any: ...
    @property
    def VISCD(self) -> Any: ...
    @property
    def VISCREF(self) -> Any: ...
    @property
    def VISDATES(self) -> Any: ...
    @property
    def VISOPTS(self) -> Any: ...
    @property
    def WAGHYSTR(self) -> Any: ...
    @property
    def WAITBAL(self) -> Any: ...
    @property
    def WALKALIN(self) -> Any: ...
    @property
    def WALQCALC(self) -> Any: ...
    @property
    def WAPI(self) -> Any: ...
    @property
    def WARN(self) -> Any: ...
    @property
    def WATDENT(self) -> Any: ...
    @property
    def WATER(self) -> Any: ...
    @property
    def WATJT(self) -> Any: ...
    @property
    def WATVISCT(self) -> Any: ...
    @property
    def WBHGLR(self) -> Any: ...
    @property
    def WBOREVOL(self) -> Any: ...
    @property
    def WCALCVAL(self) -> Any: ...
    @property
    def WCONHIST(self) -> Any: ...
    @property
    def WCONINJ(self) -> Any: ...
    @property
    def WCONINJE(self) -> Any: ...
    @property
    def WCONINJH(self) -> Any: ...
    @property
    def WCONINJP(self) -> Any: ...
    @property
    def WCONPROD(self) -> Any: ...
    @property
    def WCUTBACK(self) -> Any: ...
    @property
    def WCUTBACT(self) -> Any: ...
    @property
    def WCYCLE(self) -> Any: ...
    @property
    def WDFAC(self) -> Any: ...
    @property
    def WDFACCOR(self) -> Any: ...
    @property
    def WDRILPRI(self) -> Any: ...
    @property
    def WDRILRES(self) -> Any: ...
    @property
    def WDRILTIM(self) -> Any: ...
    @property
    def WECON(self) -> Any: ...
    @property
    def WECONINJ(self) -> Any: ...
    @property
    def WECONT(self) -> Any: ...
    @property
    def WEFAC(self) -> Any: ...
    @property
    def WELCNTL(self) -> Any: ...
    @property
    def WELDEBUG(self) -> Any: ...
    @property
    def WELDRAW(self) -> Any: ...
    @property
    def WELEVNT(self) -> Any: ...
    @property
    def WELLDIMS(self) -> Any: ...
    @property
    def WELLSTRE(self) -> Any: ...
    @property
    def WELL_COMPLETION_PROBE(self) -> Any: ...
    @property
    def WELL_PROBE(self) -> Any: ...
    @property
    def WELL_PROBE_COMP(self) -> Any: ...
    @property
    def WELL_PROBE_OPM(self) -> Any: ...
    @property
    def WELMOVEL(self) -> Any: ...
    @property
    def WELOPEN(self) -> Any: ...
    @property
    def WELOPENL(self) -> Any: ...
    @property
    def WELPI(self) -> Any: ...
    @property
    def WELPRI(self) -> Any: ...
    @property
    def WELSEGS(self) -> Any: ...
    @property
    def WELSOMIN(self) -> Any: ...
    @property
    def WELSPECL(self) -> Any: ...
    @property
    def WELSPECS(self) -> Any: ...
    @property
    def WELTARG(self) -> Any: ...
    @property
    def WELTRAJ(self) -> Any: ...
    @property
    def WFOAM(self) -> Any: ...
    @property
    def WFRICSEG(self) -> Any: ...
    @property
    def WFRICSGL(self) -> Any: ...
    @property
    def WFRICTN(self) -> Any: ...
    @property
    def WFRICTNL(self) -> Any: ...
    @property
    def WGASPROD(self) -> Any: ...
    @property
    def WGORPEN(self) -> Any: ...
    @property
    def WGRUPCON(self) -> Any: ...
    @property
    def WH2NUM(self) -> Any: ...
    @property
    def WH3NUM(self) -> Any: ...
    @property
    def WHEDREFD(self) -> Any: ...
    @property
    def WHISTCTL(self) -> Any: ...
    @property
    def WHTEMP(self) -> Any: ...
    @property
    def WINJCLN(self) -> Any: ...
    @property
    def WINJDAM(self) -> Any: ...
    @property
    def WINJFCNC(self) -> Any: ...
    @property
    def WINJGAS(self) -> Any: ...
    @property
    def WINJMULT(self) -> Any: ...
    @property
    def WINJTEMP(self) -> Any: ...
    @property
    def WLIFT(self) -> Any: ...
    @property
    def WLIFTOPT(self) -> Any: ...
    @property
    def WLIMTOL(self) -> Any: ...
    @property
    def WLIST(self) -> Any: ...
    @property
    def WLISTARG(self) -> Any: ...
    @property
    def WLISTNAM(self) -> Any: ...
    @property
    def WMICP(self) -> Any: ...
    @property
    def WNETCTRL(self) -> Any: ...
    @property
    def WNETDP(self) -> Any: ...
    @property
    def WORKLIM(self) -> Any: ...
    @property
    def WORKTHP(self) -> Any: ...
    @property
    def WPAVE(self) -> Any: ...
    @property
    def WPAVEDEP(self) -> Any: ...
    @property
    def WPIMULT(self) -> Any: ...
    @property
    def WPIMULTL(self) -> Any: ...
    @property
    def WPITAB(self) -> Any: ...
    @property
    def WPLUG(self) -> Any: ...
    @property
    def WPMITAB(self) -> Any: ...
    @property
    def WPOLYMER(self) -> Any: ...
    @property
    def WPOLYRED(self) -> Any: ...
    @property
    def WPOTCALC(self) -> Any: ...
    @property
    def WREGROUP(self) -> Any: ...
    @property
    def WRFT(self) -> Any: ...
    @property
    def WRFTPLT(self) -> Any: ...
    @property
    def WSALT(self) -> Any: ...
    @property
    def WSCCLEAN(self) -> Any: ...
    @property
    def WSCCLENL(self) -> Any: ...
    @property
    def WSCTAB(self) -> Any: ...
    @property
    def WSEED(self) -> Any: ...
    @property
    def WSEGAICD(self) -> Any: ...
    @property
    def WSEGDFIN(self) -> Any: ...
    @property
    def WSEGDFMD(self) -> Any: ...
    @property
    def WSEGDFPA(self) -> Any: ...
    @property
    def WSEGDIMS(self) -> Any: ...
    @property
    def WSEGEXSS(self) -> Any: ...
    @property
    def WSEGFLIM(self) -> Any: ...
    @property
    def WSEGFMOD(self) -> Any: ...
    @property
    def WSEGINIT(self) -> Any: ...
    @property
    def WSEGITER(self) -> Any: ...
    @property
    def WSEGLABY(self) -> Any: ...
    @property
    def WSEGLINK(self) -> Any: ...
    @property
    def WSEGMULT(self) -> Any: ...
    @property
    def WSEGPROP(self) -> Any: ...
    @property
    def WSEGSEP(self) -> Any: ...
    @property
    def WSEGSICD(self) -> Any: ...
    @property
    def WSEGSOLV(self) -> Any: ...
    @property
    def WSEGTABL(self) -> Any: ...
    @property
    def WSEGVALV(self) -> Any: ...
    @property
    def WSF(self) -> Any: ...
    @property
    def WSKPTAB(self) -> Any: ...
    @property
    def WSOLVENT(self) -> Any: ...
    @property
    def WSURFACT(self) -> Any: ...
    @property
    def WTADD(self) -> Any: ...
    @property
    def WTEMP(self) -> Any: ...
    @property
    def WTEMPQ(self) -> Any: ...
    @property
    def WTEST(self) -> Any: ...
    @property
    def WTHPMAX(self) -> Any: ...
    @property
    def WTMULT(self) -> Any: ...
    @property
    def WTRACER(self) -> Any: ...
    @property
    def WVFPDP(self) -> Any: ...
    @property
    def WVFPEXP(self) -> Any: ...
    @property
    def WWPAVE(self) -> Any: ...
    @property
    def XMF(self) -> Any: ...
    @property
    def YMF(self) -> Any: ...
    @property
    def YMODULE(self) -> Any: ...
    @property
    def ZCORN(self) -> Any: ...
    @property
    def ZFACT1(self) -> Any: ...
    @property
    def ZFACT1S(self) -> Any: ...
    @property
    def ZFACTOR(self) -> Any: ...
    @property
    def ZFACTORS(self) -> Any: ...
    @property
    def ZIPP2OFF(self) -> Any: ...
    @property
    def ZIPPY2(self) -> Any: ...
    @property
    def ZMF(self) -> Any: ...
    @property
    def ZMFVD(self) -> Any: ...

class Connection:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def attached_to_segment(self) -> bool: ...
    @property
    def center_depth(self) -> float: ...
    @property
    def cf(self) -> float: ...
    @property
    def complnum(self) -> int: ...
    @property
    def direction(self) -> str: ...
    @property
    def i(self) -> int: ...
    @property
    def j(self) -> int: ...
    @property
    def k(self) -> int: ...
    @property
    def kh(self) -> float: ...
    @property
    def number(self) -> int: ...
    @property
    def pos(self) -> Tuple[int,int,int]: ...
    @property
    def rw(self) -> float: ...
    @property
    def sat_table_id(self) -> int: ...
    @property
    def segment_number(self) -> int: ...
    @property
    def state(self) -> str: ...

class Deck:
    def __init__(self, *args, **kwargs) -> None: ...
    def active_unit_system(self, *args, **kwargs) -> Any: ...
    def add(self, keyword: DeckKeyword) -> None: ...
    def count(self, keyword: str) -> int: ...
    def default_unit_system(self, *args, **kwargs) -> Any: ...
    def __contains__(self, keyword: str) -> bool: ...
    @overload
    def __getitem__(self, index: int) -> DeckKeyword: ...
    @overload
    def __getitem__(self, keyword: str) -> DeckKeyword: ...
    @overload
    def __getitem__(self, keyword_index: tuple) -> DeckKeyword: ...
    def __iter__(self) -> Iterator: ...
    def __len__(self) -> int: ...

class DeckItem:
    def __init__(self, *args, **kwargs) -> None: ...
    def __defaulted(self, index: int) -> bool: ...
    def __has_value(self, index: int) -> bool: ...
    def __is_numeric(self) -> bool: ...
    def __uda_double(self) -> float: ...
    def __uda_str(self) -> str: ...
    def get_SI(self, index: int) -> float: ...
    def get_SI_data_list(self) -> list: ...
    def get_data_list(self) -> list: ...
    def get_int(self, index: int) -> int: ...
    def get_raw(self, index: int) -> float: ...
    def get_raw_data_list(self) -> list: ...
    def get_str(self, index: int) -> str: ...
    def get_uda(self, *args, **kwargs) -> Any: ...
    def is_double(self) -> bool: ...
    def is_int(self) -> bool: ...
    def is_string(self) -> bool: ...
    def is_uda(self) -> bool: ...
    def name(self) -> str: ...
    def __len__(self) -> int: ...

class DeckKeyword:
    @overload
    def __init__(self, parser_keyword) -> None: ...
    @overload
    def __init__(self, parser_keyword, record_list: list, active_system, default_system) -> None: ...
    @overload
    def __init__(self, parser_keyword, py_data: numpy.ndarray[numpy.int32]) -> None: ...
    @overload
    def __init__(self, parser_keyword, py_data: numpy.ndarray[numpy.float64], active_system, default_system) -> None: ...
    def get_SI_array(self) -> numpy.ndarray[numpy.float64]: ...
    def get_int_array(self) -> numpy.ndarray[numpy.int32]: ...
    def get_raw_array(self) -> numpy.ndarray[numpy.float64]: ...
    def __getitem__(self, index) -> Any: ...
    def __iter__(self) -> Iterator: ...
    def __len__(self) -> int: ...
    @property
    def name(self) -> str: ...

class DeckRecord:
    def __init__(self, *args, **kwargs) -> None: ...
    def __getitem__(self, index) -> Any: ...
    def __iter__(self) -> Iterator: ...
    def __len__(self) -> int: ...

class Dimension:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def offset(self) -> float: ...
    @property
    def scaling(self) -> float: ...

class EGrid:
    def __init__(self, filename: str, grid_name: str = ...) -> None: ...
    def active_index(self, i: int, j: int, k: int) -> int: ...
    @overload
    def cellvolumes(self) -> numpy.ndarray: ...
    @overload
    def cellvolumes(self, mask: List[int]) -> numpy.ndarray: ...
    def export_mapaxes(self) -> List[float[6]]: ...
    def global_index(self, i: int, j: int, k: int) -> int: ...
    def ijk_from_active_index(self, active_index: int) -> List[int[3]]: ...
    def ijk_from_global_index(self, global_index: int) -> List[int[3]]: ...
    @overload
    def xyz_from_active_index(self, active_index: int) -> Tuple[List[float[8]],List[float[8]],List[float[8]]]: ...
    @overload
    def xyz_from_active_index(self, active_index: int, apply_mapaxes: bool) -> Tuple[List[float[8]],List[float[8]],List[float[8]]]: ...
    @overload
    def xyz_from_ijk(self, i: int, j: int, k: int) -> Tuple[List[float[8]],List[float[8]],List[float[8]]]: ...
    @overload
    def xyz_from_ijk(self, i: int, j: int, k: int, apply_mapaxes: bool) -> Tuple[List[float[8]],List[float[8]],List[float[8]]]: ...
    @property
    def active_cells(self) -> int: ...
    @property
    def dimension(self) -> List[int[3]]: ...

class EModel:
    def __init__(self, filename: str) -> None: ...
    @overload
    def __add_filter(self, key: str, operator: str, value: int) -> None: ...
    @overload
    def __add_filter(self, key: str, operator: str, value: float) -> None: ...
    @overload
    def __add_filter(self, key: str, operator: str, value1: int, value2: int) -> None: ...
    @overload
    def __add_filter(self, key: str, operator: str, value1: float, value2: float) -> None: ...
    def active_cells(self) -> int: ...
    def active_report_step(self) -> int: ...
    def add_hc_filter(self) -> None: ...
    def get(self, key: str) -> numpy.ndarray: ...
    def get_list_of_arrays(self) -> List[Tuple[str,eclArrType]]: ...
    def get_report_steps(self) -> List[int]: ...
    def grid_dims(self) -> Tuple[int,int,int]: ...
    def has_report_step(self, rstep: int) -> bool: ...
    def reset_filter(self) -> None: ...
    def set_depth_fwl(self, fwl: List[float]) -> None: ...
    def set_report_step(self, rstep: int) -> None: ...
    def __contains__(self, parameter: str) -> bool: ...

class ERft:
    def __init__(self, filename: str) -> None: ...
    @overload
    def __get_data(self, array_name: str, well_name: str, year: int, month: int, day: int) -> Tuple[numpy.ndarray,eclArrType]: ...
    @overload
    def __get_data(self, array_name: str, report_index: int) -> Tuple[numpy.ndarray,eclArrType]: ...
    @overload
    def __get_list_of_arrays(self, report_index: int) -> List[Tuple[str,eclArrType,int]]: ...
    @overload
    def __get_list_of_arrays(self, well_name: str, year: int, month: int, day: int) -> List[Tuple[str,eclArrType,int]]: ...
    @overload
    def __has_array(self, array_name: str, report_index: int) -> bool: ...
    @overload
    def __has_array(self, array_name: str, well_name: str, date: Tuple[int,int,int]) -> bool: ...
    def __has_rft(self, well_name: str, year: int, month: int, day: int) -> bool: ...
    def __len__(self) -> int: ...
    @property
    def list_of_rfts(self) -> List[Tuple[str,Tuple[int,int,int],float]]: ...

class ERst:
    def __init__(self, filename: str) -> None: ...
    def __contains(self, tuple: Tuple[str,int]) -> bool: ...
    @overload
    def __get_data(self, index: int, report_step: int) -> Tuple[numpy.ndarray,eclArrType]: ...
    @overload
    def __get_data(self, name: str, report_step: int, occurrence: int) -> Tuple[numpy.ndarray,eclArrType]: ...
    def __has_report_step(self, report_step: int) -> bool: ...
    @overload
    def arrays(self, report_step: int) -> List[Tuple[str,eclArrType,int]]: ...
    @overload
    def arrays(self, report_step: int, lgr_name: str) -> List[Tuple[str,eclArrType,int]]: ...
    def count(self, name: str, report_step: int) -> int: ...
    def load_report_step(self, report_step: int) -> None: ...
    def __len__(self) -> int: ...
    @property
    def report_steps(self) -> List[int]: ...

class ESmry:
    def __init__(self, filename: str, load_base_run: bool = ...) -> None: ...
    def __get_all(self, key: str) -> numpy.ndarray: ...
    def __get_at_rstep(self, key: str) -> numpy.ndarray: ...
    def __start_date(self) -> Tuple[int,int,int,int,int,int,bool]: ...
    def dates(self) -> List[datetime.datetime]: ...
    @overload
    def keys(self) -> List[str]: ...
    @overload
    def keys(self, pattern: str) -> List[str]: ...
    def make_esmry_file(self) -> None: ...
    def units(self, field: str) -> str: ...
    def __contains__(self, key: str) -> bool: ...
    def __len__(self) -> int: ...

class EclFile:
    def __init__(self, filename: str, preload: bool = ...) -> None: ...
    @overload
    def __get_data(self, index: int) -> Tuple[numpy.ndarray,eclArrType]: ...
    @overload
    def __get_data(self, name: str) -> Tuple[numpy.ndarray,eclArrType]: ...
    @overload
    def __get_data(self, name: str, occurrence: int) -> Tuple[numpy.ndarray,eclArrType]: ...
    def count(self, name: str) -> int: ...
    def __contains__(self, name: str) -> bool: ...
    def __len__(self) -> int: ...
    @property
    def arrays(self) -> List[Tuple[str,eclArrType,int]]: ...

class EclOutput:
    def __init__(self, filename: str, formatted: bool = ..., append: bool = ...) -> None: ...
    def __write_c0nn_array(self, array_name: str, data: List[str], element_size: int) -> None: ...
    def __write_char_array(self, array_name: str, data: List[str]) -> None: ...
    def __write_doub_array(self, array_name: str, data: List[float]) -> None: ...
    def __write_inte_array(self, array_name: str, data: List[int]) -> None: ...
    def __write_logi_array(self, array_name: str, data: List[bool]) -> None: ...
    def __write_real_array(self, array_name: str, data: List[float]) -> None: ...
    def write_message(self, msg: str) -> None: ...

class EclipseConfig:
    def __init__(self, *args, **kwargs) -> None: ...
    def init(self, *args, **kwargs) -> Any: ...

class EclipseGrid:
    def __init__(self, *args, **kwargs) -> None: ...
    def _getXYZ(self) -> tuple: ...
    @overload
    def getCellDepth(self, g: int) -> float: ...
    @overload
    def getCellDepth(self, i: int, j: int, k: int) -> float: ...
    @overload
    def getCellDepth(self) -> numpy.ndarray: ...
    @overload
    def getCellDepth(self, mask: List[int]) -> numpy.ndarray: ...
    @overload
    def getCellVolume(self, g: int) -> float: ...
    @overload
    def getCellVolume(self, i: int, j: int, k: int) -> float: ...
    @overload
    def getCellVolume(self) -> numpy.ndarray: ...
    @overload
    def getCellVolume(self, mask: List[int]) -> numpy.ndarray: ...
    def getIJK(self, g: int) -> tuple: ...
    def globalIndex(self, i: int, j: int, k: int) -> int: ...
    @property
    def cartesianSize(self) -> int: ...
    @property
    def nactive(self) -> int: ...
    @property
    def nx(self) -> int: ...
    @property
    def ny(self) -> int: ...
    @property
    def nz(self) -> int: ...

class EclipseState:
    def __init__(self, deck: Deck) -> None: ...
    def config(self) -> EclipseConfig: ...
    def faultFaces(self, fault_name: str) -> list: ...
    def faultNames(self) -> list: ...
    def field_props(self) -> FieldProperties: ...
    def grid(self) -> EclipseGrid: ...
    def has_input_nnc(self) -> bool: ...
    def input_nnc(self) -> list: ...
    def jfunc(self) -> dict: ...
    def simulation(self) -> SimulationConfig: ...
    def tables(self) -> Tables: ...
    @property
    def title(self) -> str: ...

class FieldProperties:
    def __init__(self, *args, **kwargs) -> None: ...
    def get_double_array(self, keyword: str) -> numpy.ndarray[numpy.float64]: ...
    def get_int_array(self, keyword: str) -> numpy.ndarray[numpy.int32]: ...
    def __contains__(self, keyword: str) -> bool: ...
    def __getitem__(self, keyword: str) -> numpy.ndarray: ...

class Group:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def num_wells(self) -> int: ...
    @property
    def well_names(self) -> List[str]: ...

class IOConfig:
    def __init__(self, *args, **kwargs) -> None: ...

class InitConfig:
    def __init__(self, *args, **kwargs) -> None: ...
    def getRestartStep(self) -> int: ...
    def hasEquil(self) -> bool: ...
    def restartRequested(self) -> bool: ...

class Logger:
    def __init__(self, *args, **kwargs) -> None: ...

class OpmLog:
    def __init__(self, *args, **kwargs) -> None: ...
    def _setLogger(self, *args, **kwargs) -> Any: ...
    def bug(self, *args, **kwargs) -> Any: ...
    def debug(self, *args, **kwargs) -> Any: ...
    def error(self, *args, **kwargs) -> Any: ...
    def info(self, *args, **kwargs) -> Any: ...
    def note(self, *args, **kwargs) -> Any: ...
    def problem(self, *args, **kwargs) -> Any: ...
    def warning(self, *args, **kwargs) -> Any: ...

class ParseContext:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, action_list) -> None: ...
    def ignore_keyword(self, keyword: str) -> None: ...
    def update(self, keyword: str, action) -> None: ...

class Parser:
    def __init__(self, add_default: bool = ...) -> None: ...
    @overload
    def add_keyword(self, keyword: ParserKeyword) -> None: ...
    @overload
    def add_keyword(self, keyword_json: str) -> None: ...
    @overload
    def parse(self, filename: str) -> Deck: ...
    @overload
    def parse(self, filename: str, context: ParseContext) -> Deck: ...
    @overload
    def parse(self, filename: str, context: ParseContext, sections: List[eclSectionType]) -> Deck: ...
    @overload
    def parse_string(self, data: str) -> Deck: ...
    @overload
    def parse_string(self, data: str, context: ParseContext) -> Deck: ...
    def __getitem__(self, keyword: str) -> ParserKeyword: ...

class ParserKeyword:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def name(self) -> str: ...

class Schedule:
    def __init__(self, deck: Deck, eclipse_state) -> None: ...
    def _groups(self, report_step: int) -> List[Group]: ...
    def get_injection_properties(self, well_name: str, report_step: int) -> Dict[str,float]: ...
    def get_production_properties(self, well_name: str, report_step: int) -> Dict[str,float]: ...
    def get_well(self, well_name: str, report_step: int) -> Well: ...
    def get_wells(self, report_step: int) -> List[Well]: ...
    @overload
    def insert_keywords(self, keywords: list, step: int) -> None: ...
    @overload
    def insert_keywords(self, data: str, step: int, unit_system: UnitSystem) -> None: ...
    @overload
    def insert_keywords(self, data: str, step: int) -> None: ...
    @overload
    def insert_keywords(self, data: str) -> None: ...
    @overload
    def open_well(self, well_name: str, step: int) -> None: ...
    @overload
    def open_well(self, well_name: str) -> None: ...
    @overload
    def shut_well(self, well_name: str, step: int) -> None: ...
    @overload
    def shut_well(self, well_name: str) -> None: ...
    @overload
    def stop_well(self, well_name: str, step: int) -> None: ...
    @overload
    def stop_well(self, well_name: str) -> None: ...
    def well_names(self, well_name_pattern: str) -> List[str]: ...
    def __contains__(self, well_name: str) -> bool: ...
    def __getitem__(self, report_step: int) -> ScheduleState: ...
    def __len__(self) -> int: ...
    @property
    def end(self) -> datetime.datetime: ...
    @property
    def reportsteps(self) -> List[datetime.datetime]: ...
    @property
    def start(self) -> datetime.datetime: ...
    @property
    def timesteps(self) -> List[datetime.datetime]: ...

class ScheduleState:
    def __init__(self, *args, **kwargs) -> None: ...
    def group(self, group_name: str) -> Group: ...
    @property
    def nupcol(self) -> int: ...

class SimulationConfig:
    def __init__(self, *args, **kwargs) -> None: ...
    def hasDISGAS(self) -> bool: ...
    def hasDISGASW(self) -> bool: ...
    def hasThresholdPressure(self) -> bool: ...
    def hasVAPOIL(self) -> bool: ...
    def hasVAPWAT(self) -> bool: ...
    def useCPR(self) -> bool: ...
    def useNONNC(self) -> bool: ...

class SummaryConfig:
    def __init__(self, arg0: Deck, arg1, arg2) -> None: ...
    def __contains__(self, arg0: str) -> bool: ...

class SummaryState:
    def __init__(self, arg0: int) -> None: ...
    def elapsed(self) -> float: ...
    def group_var(self, group_name: str, variable_name: str) -> float: ...
    def has_group_var(self, group_name: str, variable_name: str) -> bool: ...
    def has_well_var(self, well_name: str, variable_name: str) -> bool: ...
    def update(self, variable_name: str, value: float) -> None: ...
    def update_group_var(self, group_name: str, variable_name: str, new_value: float) -> None: ...
    def update_well_var(self, well_name: str, variable_name: str, new_value: float) -> None: ...
    def well_var(self, well_name: str, variable_name: str) -> float: ...
    def __contains__(self, variable_name: str) -> bool: ...
    def __getitem__(self, variable_name: str) -> float: ...
    def __setitem__(self, variable_name: str, new_value: float) -> None: ...
    @property
    def groups(self) -> List[str]: ...
    @property
    def wells(self) -> List[str]: ...

class Tables:
    def __init__(self, *args, **kwargs) -> None: ...
    def evaluate(self, arg0: str, arg1: int, arg2: str, arg3: float) -> float: ...
    def __contains__(self, arg0: str) -> bool: ...

class UDAValue:
    @overload
    def __init__(self, arg0: float, arg1) -> None: ...
    @overload
    def __init__(self, arg0: str, arg1) -> None: ...
    def dimension(self, *args, **kwargs) -> Any: ...
    def get_double(self) -> float: ...
    def get_string(self) -> str: ...
    def is_double(self) -> bool: ...
    def is_string(self) -> bool: ...

class UnitSystem:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def name(self) -> str: ...

class Well:
    def __init__(self, *args, **kwargs) -> None: ...
    def available_gctrl(self) -> bool: ...
    def connections(self) -> List[Connection]: ...
    def group(self) -> str: ...
    def guide_rate(self) -> float: ...
    def isdefined(self, report_step: int) -> bool: ...
    def isinjector(self) -> bool: ...
    def isproducer(self) -> bool: ...
    def pos(self) -> Tuple[int,int,float]: ...
    def status(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def preferred_phase(self) -> str: ...

class action:
    __members__: ClassVar[dict] = ...  # read-only
    __entries: ClassVar[dict] = ...
    ignore: ClassVar[action] = ...
    throw: ClassVar[action] = ...
    warn: ClassVar[action] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class eclArrType:
    __members__: ClassVar[dict] = ...  # read-only
    C0nn: ClassVar[eclArrType] = ...
    CHAR: ClassVar[eclArrType] = ...
    DOUB: ClassVar[eclArrType] = ...
    INTE: ClassVar[eclArrType] = ...
    LOGI: ClassVar[eclArrType] = ...
    MESS: ClassVar[eclArrType] = ...
    REAL: ClassVar[eclArrType] = ...
    __entries: ClassVar[dict] = ...
    def __init__(self, value: int) -> None: ...
    def __and__(self, other: object) -> object: ...
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __gt__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __invert__(self) -> object: ...
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __or__(self, other: object) -> object: ...
    def __rand__(self, other: object) -> object: ...
    def __ror__(self, other: object) -> object: ...
    def __rxor__(self, other: object) -> object: ...
    def __setstate__(self, state: int) -> None: ...
    def __xor__(self, other: object) -> object: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class eclSectionType:
    __members__: ClassVar[dict] = ...  # read-only
    EDIT: ClassVar[eclSectionType] = ...
    GRID: ClassVar[eclSectionType] = ...
    PROPS: ClassVar[eclSectionType] = ...
    REGIONS: ClassVar[eclSectionType] = ...
    RUNSPEC: ClassVar[eclSectionType] = ...
    SCHEDULE: ClassVar[eclSectionType] = ...
    SOLUTION: ClassVar[eclSectionType] = ...
    SUMMARY: ClassVar[eclSectionType] = ...
    __entries: ClassVar[dict] = ...
    def __init__(self, value: int) -> None: ...
    def __and__(self, other: object) -> object: ...
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __gt__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __invert__(self) -> object: ...
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __or__(self, other: object) -> object: ...
    def __rand__(self, other: object) -> object: ...
    def __ror__(self, other: object) -> object: ...
    def __rxor__(self, other: object) -> object: ...
    def __setstate__(self, state: int) -> None: ...
    def __xor__(self, other: object) -> object: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

def calc_cell_vol(arg0: List[float[8]], arg1: List[float[8]], arg2: List[float[8]]) -> float: ...
def create_deck(arg0: str, arg1: ParseContext, arg2) -> Deck: ...
def create_deck_string(arg0: str, arg1: ParseContext, arg2) -> Deck: ...
