-- This reservoir simulation deck is made available under the Open Database
-- License: http://opendatacommons.org/licenses/odbl/1.0/. Any rights in
-- individual contents of the database are licensed under the Database Contents
-- License: http://opendatacommons.org/licenses/dbcl/1.0/

-- Copyright (C) 2015 Statoil

WSTAT
/

WMCTL
/

GMCTP
/

GMCTW
/

GMCTG
/

FOE

GVPR
/

WVPR
/

GVPT
/

WVPT
/

GVIR
/

WVIR
/

GVIT
/

WVIT
/

ROE
/

GWIT
/

GGIT
/

GPRW
/                              Node pressure for water injection manifolds.

ROIP_GL
/

ROIP_NL
/

RPR__GL
/

RPR__NL
/

ROPR_GL
/

ROPT_GL
/

ROPR_NL
/

ROPT_NL
/

RGPR_GL
/

RGPT_GL
/

RGPR_NL
/

RGPT_NL
/

RWPR_GL
/

RWPT_GL
/

RWPR_NL
/

RWPT_NL
/

RWIR_GL
/

RWIT_GL
/

RWIR_NL
/

RWIT_NL
/

WWIR
'F-4AH'
/

WWIT
'F-4AH'
/

WBHP
'F-4AH'
/

WTHP
'F-4AH'
/

WWIR
'I50' 'I51' 'I52' 'I55_K4'
/

WWIT
'I50' 'I51' 'I52' 'I55_K4'
/

WBHP
'I50' 'I51' 'I52' 'I55_K4'
/

WTHP
'I50' 'I51' 'I52' 'I55_K4'
/


