/*
  Copyright 2022 Equinor ASA.

  This file is part of the Open Porous Media project (OPM).

  OPM is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  OPM is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OPM.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef OPM_UTILITY_SHMATCH_HPP
#define OPM_UTILITY_SHMATCH_HPP

#include <string>

namespace Opm {

/*
  The shmatch() function is an implementation of the shell matching algorithm
  used in posix function fnmatch(). It is implemented here based on std::regex
  to avoid using a posix function.
*/

bool shmatch(const std::string& pattern, const std::string& symbol);


}
#endif //OPM_UTILITY_STRING_HPP
