/*
  Copyright (C) 2014 by Andreas Lauser

  This file is part of the Open Porous Media project (OPM).

  OPM is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  OPM is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OPM.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef OPM_PARSER_ROCKTAB_TABLE_HPP
#define	OPM_PARSER_ROCKTAB_TABLE_HPP

#include "SimpleTable.hpp"

namespace Opm {

    class DeckItem;

    class RocktabTable : public  SimpleTable {
    public:
        RocktabTable() = default;
        RocktabTable(const DeckItem& item,
                     bool isDirectional,
                     bool hasStressOption,
                     const int tableID);

        static RocktabTable serializationTestObject();

        const TableColumn& getPressureColumn() const;
        const TableColumn& getPoreVolumeMultiplierColumn() const;
        const TableColumn& getTransmissibilityMultiplierColumn() const;
        const TableColumn& getTransmissibilityMultiplierXColumn() const;
        const TableColumn& getTransmissibilityMultiplierYColumn() const;
        const TableColumn& getTransmissibilityMultiplierZColumn() const;

        bool operator==(const RocktabTable& data) const;

        template<class Serializer>
        void serializeOp(Serializer& serializer)
        {
            this->SimpleTable::serializeOp(serializer);
            serializer(m_isDirectional);
        }

    private:
        bool m_isDirectional = false;
    };
}

#endif
